import "reflect-metadata";
import {
  API,
  Characteristic,
  DynamicPlatformPlugin,
  Logger,
  PlatformAccessory,
  PlatformConfig,
  Service,
} from "homebridge";
import { ControlUnit, DeviceButton } from "./accessories";
import { Sequence } from "./accessories/sequence";
import { IConfig, IControlUnit, IDeviceButton } from "./models/config";
import { ISequence } from "./models/config/sequence";
import { PLATFORM_NAME, PLUGIN_NAME } from "./settings";
import { container } from "tsyringe";
import { HarmonyDataProvider } from "./dataProviders/harmonyDataProvider";
import { StateDataProvider } from "./dataProviders/stateDataProvider";
import { CommandService } from "./services/commandService";
import { ActivityService } from "./services/activityService";
import { VolumeService } from "./services/volumeService";

export class Platform implements DynamicPlatformPlugin {
  constructor(
    public readonly log: Logger,
    config: PlatformConfig,
    public readonly api: API
  ) {
    this.log.debug("Finished initializing platform:", config.name);
    this.config = config as unknown as IConfig;
    this.register();

    let didFinishLaunching = false;
    this.api.on("didFinishLaunching", async () => {
      log.debug("Executed didFinishLaunching callback");
      this.discoverControlUnitAccessories();
      this.discoverDeviceButtonAccessories();
      this.discoverSequenceAccessories();
      this.pruneAccessories();
      didFinishLaunching = true;
    });
  }

  public readonly Service: typeof Service = this.api.hap.Service;
  public readonly Characteristic: typeof Characteristic =
    this.api.hap.Characteristic;

  // this is used to track restored cached accessories
  public readonly accessories: PlatformAccessory[] = [];
  public config: IConfig;

  /**
   * Discover new control unit accessories
   * @param dataProvider
   */
  private discoverControlUnitAccessories(): void {
    this.config.ControlUnits.forEach((unit: IControlUnit) => {
      const uuid = this.api.hap.uuid.generate(unit.DisplayName);
      const existingAccessory = this.accessories.find((e) => e.UUID === uuid);

      if (existingAccessory) {
        this.log.info(
          "Restoring existing accessory from cache: " +
            existingAccessory.displayName
        );

        new ControlUnit(
          this,
          existingAccessory,
          container.resolve(ActivityService),
          container.resolve(CommandService),
          container.resolve(VolumeService),
          unit.Activities
        );

        this.api.publishExternalAccessories(PLUGIN_NAME, [existingAccessory]);
        console.log("Publishing external accessory: " + uuid);
      } else {
        this.log.info("Adding new accessory: " + unit.DisplayName);
        const accessory = new this.api.platformAccessory(
          unit.DisplayName,
          uuid
        );
        accessory.context["DeviceName"] = unit.DisplayName;
        accessory.context["Type"] = typeof ControlUnit.name;

        new ControlUnit(
          this,
          accessory,
          container.resolve(ActivityService),
          container.resolve(CommandService),
          container.resolve(VolumeService),
          unit.Activities
        );

        this.api.publishExternalAccessories(PLUGIN_NAME, [accessory]);
        console.log("Publishing external accessory: " + uuid);
      }
    });
  }

  /**
   * Discover new device button accessories
   * @param dataProvider
   */
  private discoverDeviceButtonAccessories(): void {
    this.config.DeviceButtons.forEach((button: IDeviceButton) => {
      const uuid = this.api.hap.uuid.generate(button.DisplayName);
      const existingAccessory = this.accessories.find((e) => e.UUID === uuid);
      if (existingAccessory) {
        this.log.info(
          "Restoring existing accessory from cache: " +
            existingAccessory.displayName
        );

        new DeviceButton(
          this,
          existingAccessory,
          container.resolve(HarmonyDataProvider),
          button
        );
        this.api.updatePlatformAccessories([existingAccessory]);
      } else {
        this.log.info("Adding new accessory: " + button.DisplayName);
        const accessory = new this.api.platformAccessory(
          button.DisplayName,
          uuid
        );
        accessory.context["DeviceName"] = button.DisplayName;
        accessory.context["Type"] = typeof DeviceButton;

        new DeviceButton(
          this,
          accessory,
          container.resolve(HarmonyDataProvider),
          button
        );

        this.api.registerPlatformAccessories(PLUGIN_NAME, PLATFORM_NAME, [
          accessory,
        ]);
        this.accessories.push(accessory);
      }
    });
  }

  /**
   * Discover new sequence accessories
   * @param dataProvider
   */
  private discoverSequenceAccessories(): void {
    this.config.Sequences.forEach((sequence: ISequence) => {
      const uuid = this.api.hap.uuid.generate(sequence.DisplayName);
      const existingAccessory = this.accessories.find((e) => e.UUID === uuid);
      if (existingAccessory) {
        this.log.info(
          "Restoring existing accessory from cache: " +
            existingAccessory.displayName
        );

        new Sequence(
          this,
          existingAccessory,
          container.resolve(HarmonyDataProvider),
          sequence
        );
        this.api.updatePlatformAccessories([existingAccessory]);
      } else {
        this.log.info("Adding new accessory: " + sequence.DisplayName);
        const accessory = new this.api.platformAccessory(
          sequence.DisplayName,
          uuid
        );
        accessory.context["DeviceName"] = sequence.DisplayName;
        accessory.context["Type"] = typeof "Sequence";

        new Sequence(
          this,
          accessory,
          container.resolve(HarmonyDataProvider),
          sequence
        );

        this.api.registerPlatformAccessories(PLUGIN_NAME, PLATFORM_NAME, [
          accessory,
        ]);
        this.accessories.push(accessory);
      }
    });
  }

  private pruneAccessories(): void {
    // Remove devices not discovered
    const listOfConfiguredAccessories = [
      ...this.config.DeviceButtons.map((e) => e.DisplayName),
      ...this.config.Sequences.map((e) => e.DisplayName),
      ...this.config.ControlUnits.map((e) => e.DisplayName),
    ];
    for (const accessory of this.accessories) {
      if (
        listOfConfiguredAccessories.filter(
          (displayName) => displayName === accessory.displayName
        ).length === 0
      ) {
        this.api.unregisterPlatformAccessories(PLUGIN_NAME, PLATFORM_NAME, [
          accessory,
        ]);
      }
    }
  }

  configureAccessory(accessory: PlatformAccessory<Record<string, any>>): void {
    this.log.info("Loading accessory from cache:", accessory.displayName);

    // add the restored accessory to the accessories cache so we can track if it has already been registered
    this.accessories.push(accessory);
  }

  private register(): void {
    container.register<Platform>(Platform, { useValue: this });
    container.register<IConfig>("IConfig", { useValue: this.config });
    container.register<Logger>("log", { useValue: this.log });
    container.registerSingleton<HarmonyDataProvider>(HarmonyDataProvider);
    container.registerSingleton<ActivityService>(ActivityService);
    container.registerSingleton<StateDataProvider>(StateDataProvider);
    container.registerSingleton<CommandService>(CommandService);
  }
}
